import os
import sys
import glob
import subprocess

LOG_FILE_EXTENSION = '*.txt'
BASH_DIR = 'Bash'
BASH_EXE_NAME = 'bash.exe'
FLASH_BASH = 'flash.bash'
M4_SUBFILE = 'm4tmp.bin'

def main(argv=None):
    arglen = len(sys.argv)

    if (arglen != 7):
        print("missing argument to jlink flash script")
        return 1

    file_name = sys.argv[1]
    file_size = sys.argv[2]
    file_offset = sys.argv[3]
    jlink_path = sys.argv[4]
    file_tag = sys.argv[5]
    verify_flag = sys.argv[6]

    # add JLink path to PATH env. variable and export JLINKCMD variable for flash bash script.
    os.environ["PATH"] += os.pathsep + jlink_path
    os.environ["JLINKCMD"] = 'JLink.exe'

    # remove all log files first.
    filelist = glob.glob(os.path.join(os.getcwd(), LOG_FILE_EXTENSION))
    for f in filelist:
        os.remove(f)

    # check for bash executable.
    parentdir = os.path.normpath(os.getcwd()+os.sep+os.pardir)
    bashfile = os.path.join(parentdir, BASH_DIR, BASH_EXE_NAME)
    if not os.path.isfile(bashfile):
        print ("Error: cannot find bash!")
        return 1

    # form the commandline arguments for flashing
    cmd_args = [bashfile, FLASH_BASH, file_name, file_size, file_offset, verify_flag]

    # invoke bash
    try:
        process_status = subprocess.call(cmd_args, stderr=subprocess.DEVNULL)
    except Exception as e:
        print(e)
        return 1

    if (process_status != 0):
        return 1

    # if M4, write a sub file with a known pattern to indicate M4-image is flashed
    if (file_tag == 'M4APP'):
        # remove the sub-file if it already exist
        if os.path.isfile(M4_SUBFILE):
            os.remove(M4_SUBFILE)

        f = open(M4_SUBFILE, 'w')
        f.write('1')
        f.close()
        sub_args = [bashfile, FLASH_BASH, M4_SUBFILE, '0x1002', '0x1C000']

        try:
            sub_status = subprocess.call(sub_args, stderr=subprocess.DEVNULL)
        except Exception as e:
            print(e)
            return 1

        #remove the sub-file
        if os.path.isfile(M4_SUBFILE):
            os.remove(M4_SUBFILE)

        return sub_status

    return 0

if __name__ == "__main__":
    sys.exit(main())


