'''
Created on Jun 16, 2018
@author: coolbreeze


██╗  ██╗██╗   ██╗███╗   ███╗ ██████╗ ██████╗ ███████╗███╗   ███╗
╚██╗██╔╝╚██╗ ██╔╝████╗ ████║██╔═══██╗██╔══██╗██╔════╝████╗ ████║
 ╚███╔╝  ╚████╔╝ ██╔████╔██║██║   ██║██║  ██║█████╗  ██╔████╔██║
 ██╔██╗   ╚██╔╝  ██║╚██╔╝██║██║   ██║██║  ██║██╔══╝  ██║╚██╔╝██║
██╔╝ ██╗   ██║   ██║ ╚═╝ ██║╚██████╔╝██████╔╝███████╗██║ ╚═╝ ██║
╚═╝  ╚═╝   ╚═╝   ╚═╝     ╚═╝ ╚═════╝ ╚═════╝ ╚══════╝╚═╝     ╚═╝

#######################    PUBLIC DOMAIN SOFTWARE    #######################
# This is free and unencumbered software released into the public domain.
# 
# Anyone is free to copy, modify, publish, use, compile, sell, or
# distribute this software, either in source code form or as a compiled
# binary, for any purpose, commercial or non-commercial, and by any
# means.
# 
# In jurisdictions that recognize copyright laws, the author or authors
# of this software dedicate any and all copyright interest in the
# software to the public domain. We make this dedication for the benefit
# of the public at large and to the detriment of our heirs and
# successors. We intend this dedication to be an overt act of
# relinquishment in perpetuity of all present and future rights to this
# software under copyright law.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
# For more information, please refer to <http://unlicense.org/>
#######################    PUBLIC DOMAIN SOFTWARE    #######################

Thanks:
(1) http://patorjk.com/software/taag/#p=display&f=ANSI%20Shadow&t=xymodem%0A
(2) https://fieldeffect.info/wp/frontpage/xmodem/
(3) http://www.kiv.zcu.cz/~luki/vyuka/stare-materialy/pd/ymodem.html

'''


import logging
import os
import stat
import platform
import binascii
import time
import serial
import argparse
import sys
from serial.serialutil import SerialException
import traceback
import signal
import datetime


# TODO
# 1. return values for all
# 2. timeout for serial read and serial write handling
# 3. eliminate receive block zero function
# 4. if timeout, flush buffer (input) and try again
# 5. while waiting for ACK/NAK of block(receive_block), if getting unknown bytes, flush input buffer and then send NCG again
# 6. while waiting for NCG byte, if getting unknown bytes, flush input buffer and wait for NCG again


class XYmodem:
    
    crc16table = [
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
    ]
    
    # constants for the modem protocols
    PACKET_SIZE_128             = 128
    PACKET_SIZE_1024            = 1024
    
    PROTOCOL_INVALID            = 0
    PROTOCOL_XMODEM             = 1
    PROTOCOL_XMODEM1k           = 2
    PROTOCOL_YMODEM             = 3
    
    CANCEL                      = b'\x18'
    NAK                         = b'\x15'
    CRC                         = b'C'
    SOH                         = b'\x01'
    STX                         = b'\x02'
    EOT                         = b'\x04'
    ACK                         = b'\x06'
    PAD                         = b'\x1a'
    NUL                         = b'\x00'
    SPACE                       = b'\x20'  
    
    # return values of tx__wait_for_ncg()
    NCG_NAK                     = 1
    NCG_CRC                     = 2
    NCG_CANCEL                  = 3
    NCG_FAIL                    = 4
    
    # return values of tx__wait_for_response()
    RESP_NAK                    = 1
    RESP_ACK                    = 2
    RESP_CANCEL                 = 3
    RESP_UNKNOWN                = 4
    
    
    # return values of tx mode send functions: send_block, send_block_zero, send_file, send_files
    TX_SEND_OK                  = 0
    TX_SEND_FAIL                = 1
    # we can specialize the FAIL values further if required, but probably overkill
    # use the debug logs to figure out is a better way
    
    
    # return values of rx__wait_for_hdr()
    DATA_SOH                    = 1
    DATA_STX                    = 2
    DATA_CANCEL                 = 3
    DATA_EOT                    = 4
    DATA_UNKNOWN                = 5
    
    # return values of rx mode send functions: receive_block, receive_block_zero, receive_file, receive_files
    RX_RECV_OK                  = 0
    RX_RECV_FAIL                = 1
    # we can specialize the FAIL values further if required, but probably overkill
    # use the debug logs to figure out is a better way
    
    
    # return values of receiver_receive_block()
    RECV_OK                     = 0
    RECV_FAIL                   = 1 
    RECV_EOT                    = 2
    
    # return values of send_files()/receive_files()/ - used by clients
    STATUS_OK                   = 0
    STATUS_FAIL                 = 1
    
    # default values
    RETRY_COUNT_MAX_DEFAULT     = 10
    CANCEL_COUNT_MAX_DEFAULT    = 2
    TIMEOUT = 2
    
    
    def __init__(self, 
                 protocol, 
                 transport_stream, 
                 crc_enabled,
                 retry_count_max = RETRY_COUNT_MAX_DEFAULT, 
                 cancel_count_max = CANCEL_COUNT_MAX_DEFAULT):
        
        self.protocol = protocol
        self.transport_stream = transport_stream
        self.crc_enabled = crc_enabled
        self.retry_count_max = retry_count_max
        self.cancel_count_max = cancel_count_max
        
        
        logging.debug("%-24s [%s]", "protocol", self.protocol)
        logging.debug("%-24s [%s]", "transport_stream", self.transport_stream)
        logging.debug("%-24s [%s]", "crc_enabled", self.crc_enabled)
        logging.debug("%-24s [%d]", "retry_count_max", self.retry_count_max)
        logging.debug("%-24s [%d]", "cancel_count_max", self.cancel_count_max)



    ###################################        SENDER FUNCTIONS ++        ###################################
    
    # sender:
    # send_files(file_list)                    [common]
    #    send_block_zero(file_path)            [ymodem - assimilate file details to send]
    #    send_file(file_path)                  [common]
    #        wait_for_ncg()                    [wait for receiver NCG (NAK/CRC/CANCEL/UNKNOWN), in send_block_zero(), send_file()]
    #        send_block()                      [send block content, wait for response, decide based on response]
    #            wait_for_response()           [after sending block, wait for response (ACK, NACK, CANCEL, UNKNOWN, EOT)]
    #            retry-logic in send_block
    #        send_eot()                        [execute eot sequence] 
    # create_header()                          
    # create_footer()   

    def tx__wait_for_ncg(self):
        # wait for the receiver to give a go ahead
        logging.debug("wait for NCG")
        
        cancel_count = 0
        error_count = 0
        
        # START of wait for ncg while loop
        while(True):
                                
            ncgbyte = self.transport_stream.read(1)
            
            if (ncgbyte == XYmodem.NAK):                
                # receiver expects checksum
                logging.debug("NCGByte = NAK")                
                return XYmodem.NCG_NAK
            
            elif (ncgbyte == XYmodem.CRC):                                
                # receiver expects CRC16
                logging.debug("NCGByte = CRC")
                return XYmodem.NCG_CRC
            
            elif (ncgbyte == XYmodem.CANCEL):                
                cancel_count += 1
                logging.debug("%s %d", "NCGByte = CANCEL, count =", cancel_count)
                
                if(cancel_count >= self.cancel_count_max):                    
                    logging.debug("receiver aborted transfer")                
                    return XYmodem.NCG_CANCEL
                
                else:                                    
                    continue # wait for more CANCELs
                
            else:
                if(len(ncgbyte) == 0):
                    logging.debug("timed out waiting for NCG")
                    
                else:
                    # we are getting unexpected bytes from receiver. we purge stuff from input and wait again.
                    logging.debug("%s %s, %s = %d", "unknown NCGByte =", ncgbyte, "error_count", error_count)
                    self.purge_transport_stream_input(sleep_interval_sec = 5)                
                    logging.debug("purged input buffer")                      
                
                error_count += 1                
                
                if(error_count >= self.retry_count_max):
                    logging.error("retry count exceeded threshold, abort")
                    return XYmodem.NCG_FAIL
                
                else:
                    logging.debug("waiting again for correct NCG")
                    continue # wait for further bytes in case a proper NCG Byte is received
                    
        # END of wait for ncg while loop
        
    
    def tx__wait_for_response(self):
        # wait for ACK/NAK/CANCEL
        logging.debug("wait for response")
        
        cancel_count = 0
        error_count = 0
        
        while(True):
            
            resp_byte = self.transport_stream.read(1)
            
            if(resp_byte == XYmodem.ACK):                
                logging.debug("got ACK")
                return XYmodem.RESP_ACK
            
            elif(resp_byte == XYmodem.NAK):                
                logging.debug("got NAK")
                return XYmodem.RESP_NAK                
            
            elif(resp_byte == XYmodem.CANCEL):                
                cancel_count += 1
                logging.debug("%s [%d]", "got CANCEL", cancel_count)
                
                if(cancel_count >= XYmodem.cancel_count_max):                    
                    logging.debug("receiver cancelled")
                    return XYmodem.RESP_CANCEL
                
                else:                    
                    continue # wait for more CANCELs
            
            else:
                error_count += 1    
                
                if(len(resp_byte) == 0):
                    logging.debug("timed out waiting for response")
                    
                else:
                    # we are getting unexpected bytes from receiver. do we flush the input here ?
                    logging.debug("%s: [%s], %s: %d","got unkown response", resp_byte, "error_count", error_count)
                    self.purge_transport_stream_input(sleep_interval_sec = 5)
                    logging.debug("purged input buffer")                    
                
                if(error_count >= self.retry_count_max):                    
                    logging.error("retry count exceeded threshold, abort")
                    return XYmodem.RESP_UNKNOWN
                
                else:
                    logging.debug("waiting again for correct response")                    
                    continue # wait for further bytes (retry) in case a proper response byte is received

    
    # add return value
    def tx__send_eot(self):
        # EOT:
        logging.debug("send EOT 1")
        send_buffer = []
        send_buffer.append(ord(XYmodem.EOT))
        self.transport_stream.write(send_buffer)
        
        # wait for ACK or NAK
        logging.debug("wait for receiver response")
        resp_byte = self.transport_stream.read(1)
        if(resp_byte == XYmodem.NAK):
            logging.debug("resp = NAK")
            logging.debug("send EOT 2")
            
            send_buffer = []
            send_buffer.append(ord(XYmodem.EOT))
            self.transport_stream.write(send_buffer)
            logging.debug("wait for receiver response")
            resp_byte = self.transport_stream.read(1)
            if(resp_byte == XYmodem.ACK):
                logging.debug("resp = ACK")
                
            elif(resp_byte == XYmodem.NAK):
                logging.debug("resp = NAK")
                
            else:
                logging.debug("unexpected: " + str(resp_byte))
                
        elif(resp_byte == XYmodem.ACK):
            logging.debug("resp = ACK")
            
        else:
            logging.debug("unexpected: " + str(resp_byte))
            
        return

    
    def tx__create_header(self, header_buffer, packet_size, sequence_index):
        # SOH/STX:
        if(packet_size == XYmodem.PACKET_SIZE_128):                    
            header_buffer.append(ord(XYmodem.SOH))
            
        elif(packet_size == XYmodem.PACKET_SIZE_1024):            
            header_buffer.append(ord(XYmodem.STX))
            
        # PACKET NUMBER:
        sequence_num = sequence_index % 256            
        header_buffer.append(sequence_num)
        
        # PACKET NUMBER 1's complement
        header_buffer.append((255 - sequence_num) % 256)
        
        return
    
        
    def tx__create_footer(self, footer_buffer, block_data):
        
        # CHECKSUM/CRC:
        if(self.crc_enabled == True):
            crc = 0
            crc = self.calc_crc16(block_data)
            footer_buffer.extend([crc >> 8, crc & 0xff])
            
        else:                
            checksum = 0
            if platform.python_version_tuple() >= ('3', '0', '0'):
                checksum = (sum(block_data) + checksum) % 256
                
            else:
                checksum = (sum(map(ord, block_data)) + checksum) % 256
                
            footer_buffer.append(checksum)
        
        return      
    
    
    def tx__send_block(self, block_data, packet_size, sequence_index):
        
        # block_data is assumed to be of size specified in packet_size already
        # padding to be done by caller if needed, before calling send!
        # create the header, footer
        header_buffer = bytearray([])
        self.tx__create_header(header_buffer = header_buffer, 
                               packet_size = packet_size, 
                               sequence_index = sequence_index)
        
        footer_buffer = bytearray([])
        self.tx__create_footer(footer_buffer = footer_buffer, 
                               block_data = block_data)
        
        retry_count = 0
        while(True):
            # send the packet
            self.transport_stream.write(header_buffer)
            self.transport_stream.write(block_data)
            self.transport_stream.write(footer_buffer)
            logging.debug("sending block %d [%d bytes]", sequence_index, len(block_data))
            
            # wait for response: ACK/NACK/CANCEL
            response = self.tx__wait_for_response()
            
            if(response == XYmodem.RESP_ACK):                
                # ok, proceed
                logging.debug("block %d sent OK [%d bytes]", sequence_index, len(block_data))
                return XYmodem.TX_SEND_OK
            
            elif(response == XYmodem.RESP_NAK):
                # receiver indicated that received block data is corrupted
                # we will try to send the same block again upto retry_count_max                
                retry_count += 1
                 
                if(retry_count >= self.retry_count_max):                    
                    # abort, failed to send block
                    logging.debug("block %d send FAILED [NAK]", sequence_index)
                    return XYmodem.TX_SEND_FAIL
                
                else:                    
                    # try sending the block again
                    continue
                
            elif(response == XYmodem.RESP_CANCEL):                
                # receiver wants to abort transfer
                logging.debug("block %d send FAILED [CANCEL]", sequence_index)
                return XYmodem.TX_SEND_FAIL
            
            elif(response == XYmodem.RESP_UNKNOWN):                
                # we are getting unexpected(garbage) responses from receiver, abort
                logging.debug("block %d send FAILED [UNKNOWN/TIMEDOUT]", sequence_index)
                return XYmodem.TX_SEND_FAIL
        
        
    
    def tx__send_block_zero(self, file_path):
        
        # STEP 1: gather details to fill in block zero
        if(file_path == None):
            # no special data to fill in, we will pad with NULs
            logging.debug("block zero with NUL sequence (end of batch)")
        
        else:
            logging.debug("block zero with file details (next file in batch)")
            if(not os.path.exists(file_path)):
                logging.error("file [%s] does not exist, abort", file_path)
                return XYmodem.TX_SEND_FAIL            
            
            file_name = os.path.basename(file_path)
            file_stats = os.stat(file_path)    
            file_size = file_stats[stat.ST_SIZE]
            if(file_size == 0):
                logging.error("file [%s] size is 0 bytes, abort", file_path)
                return XYmodem.TX_SEND_FAIL
            
            # modified time is for future use, send if required only.
            #file_mtime_epoch = file_stats[stat.ST_MTIME]
            #file_mtime_readable = time.strftime("%H:%M:%S %a %b %d %Y", time.localtime(file_mtime_epoch))
            
            #logging.debug("%-24s: %s", "file_path", file_path)
            #logging.debug("%-24s: %s", "file name", file_name)     
            #logging.debug("%-24s: %s", "file size", file_size)    
            #logging.debug("%-24s: %s", "file time (epoch)", file_mtime_epoch)
            #logging.debug("%-24s: %s", "file time (human)", file_mtime_readable)
            #logging.debug("")
            # converting to bytes using ascii 
            #logging.debug("%-24s: %s", "file name (HEX)", binascii.hexlify(str(file_name).encode("ascii")))
            #logging.debug("%-24s: %s", "file size (HEX)", binascii.hexlify(str(file_size).encode("ascii")))
            #logging.debug("%-24s: %s", "file time epoch (HEX)", binascii.hexlify(str(file_mtime_epoch).encode("ascii")))
            #logging.debug("%-24s: %s", "file time human (HEX)", binascii.hexlify(str(file_mtime_readable).encode("ascii")))
            #logging.debug("")
        
        
        # STEP 2: Wait For Receiver's NCG : CRC/NAK
        ncgresponse = self.tx__wait_for_ncg()
        
        if(ncgresponse == XYmodem.NCG_CRC):            
            if(self.crc_enabled == False):                
                logging.warning("switching from CHECKSUM to CRC, as receiver expects CRC only")
                
            self.crc_enabled = True
            
        elif(ncgresponse == XYmodem.NCG_NAK):            
            if(self.crc_enabled == True):                
                logging.warning("switching from CRC to CHECKSUM, as receiver expects CHECKSUM only")
                
            self.crc_enabled = False
            
        else:            
            # either we got xyzmodem_NCG_CANCEL or xyzmodem_NCG_UNKNOWN -> abort the send
            return XYmodem.TX_SEND_FAIL
    
        
        # STEP 3: send file info block (block zero)
        block_data = bytearray([])
        
        if(file_path == None):
            # this is the NUL sequence (end of batch), no data to add!
            # \x00...\x00
            packet_size = XYmodem.PACKET_SIZE_128
            
        else:
            # this is the file info sequence, add the file data
            # NAME\x00SIZE(DEC)\x20MODTIME(OCT)\x20MODE(OCT)\x20\x00...\x00
            file_name_bytes = file_name.encode("ascii")
            file_size_bytes = str(file_stats[stat.ST_SIZE]).encode("ascii")
            
            packet_size = 0
            max_file_name_len = 0
            if( (len(file_name_bytes) + 1 + len(file_size_bytes) + 1) > XYmodem.PACKET_SIZE_128):                
                max_file_name_len = XYmodem.PACKET_SIZE_1024 - 1 - len(file_size_bytes) - 1
                packet_size = XYmodem.PACKET_SIZE_1024
                                
            else:                
                max_file_name_len = XYmodem.PACKET_SIZE_128 - 1 - len(file_size_bytes) - 1
                packet_size = XYmodem.PACKET_SIZE_128
                
            block_data.extend(file_name_bytes[:max_file_name_len])         # filename , truncate in case too long.
            block_data.append(ord(XYmodem.NUL))                            # NULL
            block_data.extend(file_size_bytes)                             # filesize
            block_data.append(ord(XYmodem.SPACE))                          # \x20 (space)
            
        # adjust the data bytes to packet size length by filling in NULs
        block_data = block_data.ljust(packet_size, XYmodem.NUL)
        
        # send block zero:
        sequence_index = 0                                                 # zero
        tx_status = self.tx__send_block(block_data = block_data, 
                                        packet_size = packet_size, 
                                        sequence_index = sequence_index)
        
        return tx_status

    
    # add return values
    def tx__send_file(self, file_path):
        # STEP 1: init info for sending file
    
        # basic error checks:
        if(self.transport_stream == None):            
            logging.error("transport_stream not specified")
            return XYmodem.TX_SEND_FAIL
        
        if(self.protocol == None):            
            logging.error("protocol not specified")
            return XYmodem.TX_SEND_FAIL
        
        data_stream = None
        
        try:            
            data_stream = open(file_path, 'rb')
            
        except OSError as oE:            
            logging.error("could not open file:", file_path, oE)
            return XYmodem.TX_SEND_FAIL
        
        file_name = os.path.basename(file_path)
        file_stats = os.stat(file_path)    
        file_size = file_stats[stat.ST_SIZE]
                
        logging.debug("%-24s: %s", "file_path", file_path)
        logging.debug("%-24s: %s", "file name", file_name)     
        logging.debug("%-24s: %s", "file size", file_size)    
        logging.debug("")
        # converting to bytes using ascii 
        logging.debug("%-24s: %s", "file name (HEX)", binascii.hexlify(str(file_name).encode("ascii")))
        logging.debug("%-24s: %s", "file size (HEX)", binascii.hexlify(str(file_size).encode("ascii")))
        logging.debug("")
        
        
        # STEP 2: Wait For Receiver's NCG
        ncgresponse = self.tx__wait_for_ncg()
        
        if(ncgresponse == XYmodem.NCG_CRC):            
            if(self.crc_enabled == False):
                logging.warning("switching from CHECKSUM to CRC, as receiver expects CRC only")
            
            self.crc_enabled = True
            
        elif(ncgresponse == XYmodem.NCG_NAK):            
            if(self.crc_enabled == True):
                logging.warning("switching from CRC to CHECKSUM, as receiver expects CHECKSUM only")
            
            self.crc_enabled = False
            
        else:            
            # xyzmodem_NCG_CANCEL, xyzmodem_NCG_UNKNOWN -> abort the send
            return XYmodem.TX_SEND_FAIL
        
        
        # STEP 3: Send File Content Loop
        total_bytes_sent = 0
        sequence_index = 1  # start from 1    
        
        # START of send file while loop
        while(total_bytes_sent != file_size):
        
            block_data = bytearray([])
                
            # read in next packet of data:
            packet_size = 0
            
            if(self.protocol == XYmodem.PROTOCOL_XMODEM):                
                # always send in 128 B packets for xmodem
                packet_size = XYmodem.PACKET_SIZE_128
                
            else:                
                # for xmodem1k/ymodem, send 128 B or 1024 B packet
                # according to remaining file_size
                remaining_file_size = file_size - total_bytes_sent
                
                if(remaining_file_size >= XYmodem.PACKET_SIZE_1024):
                    packet_size = XYmodem.PACKET_SIZE_1024                    
                
                else:
                    packet_size = XYmodem.PACKET_SIZE_128
            
            
            # read in the data from the file
            block_data = data_stream.read(packet_size)
            total_bytes_sent += len(block_data)
            
            # pad the block_data if needed
            block_data = block_data.ljust(packet_size, XYmodem.PAD)
            
            # send the block
            tx_status = self.tx__send_block(block_data = block_data, 
                                            packet_size = packet_size, 
                                            sequence_index = sequence_index)
            
            if(tx_status == XYmodem.TX_SEND_FAIL):
                logging.error("send block [%d] failed, abort", sequence_index)
                return XYmodem.TX_SEND_FAIL 
            
            # send ok, move on to the next block
            sequence_index += 1                
                   
        # END of send file while loop
        
        
        # STEP 4: End Of Transfer Sequence
        self.tx__send_eot()
        
        # DONE sending current file
        return XYmodem.TX_SEND_OK
    
    
    def send(self, file_list):
        # START of Process File List while loop
        file_number = 1
        for file_path in file_list:
            
            logging.debug("processing: [%d] %s", file_number, file_path)
            
            # does file exist ?
            if(not os.path.isfile(file_path)):                
                logging.warning("no such file, skipping: ", file_path)
                file_number += 1                
                continue
            
            # send block zero for file (ymodem)
            if(self.protocol == XYmodem.PROTOCOL_YMODEM):                
                # execute the ymodem block zero process here indicating file details
                logging.debug("send block zero for: %s", file_path)
                
                if(self.tx__send_block_zero(file_path) == XYmodem.TX_SEND_FAIL):
                    logging.error("send block zero for: %s failed", file_path)
                    return XYmodem.TX_SEND_FAIL
            
            # send file content (common)
            send_status = self.tx__send_file(file_path)
            
            # send ok?
            if(send_status == XYmodem.TX_SEND_FAIL):
                logging.error("send file: %s failed", file_path)
                return XYmodem.TX_SEND_FAIL
            
            # next file
            file_number += 1    
        
        # END of Process File List while loop
        
        # send block zero indicating all files in batch are done (ymodem)
        if(self.protocol == XYmodem.PROTOCOL_YMODEM):            
            # execute the ymodem block zero process here indicating no further files
            logging.debug("end of batch")
            
            if(self.tx__send_block_zero(None) == XYmodem.TX_SEND_FAIL):
                logging.error("send block zero for end of batch failed")
                return XYmodem.TX_SEND_FAIL
            
        # all ok, we end up here
        return XYmodem.TX_SEND_OK
    ###################################        SENDER FUNCTIONS --        ###################################


    ###################################        RECEIVER FUNCTIONS ++        ###################################
    
    # receiver:
    # receive_files(dir_path)                  [common]
    #    receive_block_zero()                  [ymodem - send NCG, receive block, process file details]
    #    receive_file(dir_path, file_name)     [common]
    #        send_NCG                          [send NAK/CRC based on crc enabled]  
    #        receive_block()                   [wait for hdr(SOH/STX/CANCEL/UNKNOWN), process data, send ACK/NAK/CANCEL]
    #        wait_for_hdr()                    [wait for (SOH/STX/CANCEL/UKNOWN handle cancel count, unknown count]

    def rx__do_ncg_handshake(self):
        retry_count = 0
        
        while(True):
            # before sending the NCG to initiate the transfer, flush any input in the transport stream
            self.purge_transport_stream_input(sleep_interval_sec = 0)     
            
            # send NCG Byte : CRC or NAK
            if(self.crc_enabled == True):
                self.transport_stream.write(XYmodem.CRC)        
                    
            else:
                self.transport_stream.write(XYmodem.NAK)
            
            # wait for sender to send valid data    
            sender_response = self.rx__wait_for_hdr()
            
            if((sender_response == XYmodem.DATA_UNKNOWN) or
                sender_response == XYmodem.DATA_EOT):
                retry_count += 1
                
                if(retry_count >= self.retry_count_max):                    
                    logging.debug("exhausted retry max threshold, giving up")
                    return XYmodem.DATA_UNKNOWN
                
                else:
                    # send the NCG byte again, and wait for sender
                    continue
                
            else:
                return sender_response
        
    
    def rx__wait_for_hdr(self):
        cancel_count = 0
        error_count = 0
        
        logging.debug("wait for sender")
            
        # START of wait-for-sender-loop
        while(True):
            data_byte = self.transport_stream.read(1) # SOH/STX or EOT or CANCEL or UNKNOWN/TIMEOUT
            
            if(data_byte == XYmodem.SOH):                
                logging.debug("got SOH")
                return XYmodem.DATA_SOH
            
            elif(data_byte == XYmodem.STX):                
                logging.debug("got STX")
                return XYmodem.DATA_STX
            
            elif(data_byte == XYmodem.EOT):                
                logging.debug("got EOT")
                return XYmodem.DATA_EOT
            
            elif(data_byte == XYmodem.CANCEL):                
                cancel_count += 1
                logging.debug("%s [%d]", "got CANCEL", cancel_count)
                
                if(cancel_count >= self.cancel_count_max):                    
                    logging.debug("sender cancelled")
                    return XYmodem.DATA_CANCEL
                
                else:                    
                    continue # wait for more CANCELs
            
            else: # unknown byte or timeout happened!                
                error_count += 1
                
                if(len(data_byte) == 0):
                    logging.debug("timed out waiting for sender, error_count: %d",error_count)
                    
                else:
                    logging.debug("%s [%s] %s = %d", "got unknown byte", data_byte, "error_count", error_count)
                    # as we are getting unexpected bytes, purge all input and try again.
                    self.purge_transport_stream_input(sleep_interval_sec = 2)
                    logging.debug("purged input buffer")
                
                if(error_count >= self.retry_count_max):                    
                    logging.debug("exhausted retry max threshold, giving up")
                    return XYmodem.DATA_UNKNOWN
                
                else:
                    # continue to wait for sender                    
                    continue
        # END of wait-for-sender-loop

    
    def rx__eot_sequence(self):
        
        # execute the EOT sequence, specific to xmodem/ymodem
        if( (self.protocol == XYmodem.PROTOCOL_XMODEM) or
            (self.protocol == XYmodem.PROTOCOL_XMODEM1k) ):                    
            logging.debug("[XMODEM] sending ACK")
            # xmodem senders expect an ACK on EOT 1
            self.transport_stream.write(XYmodem.ACK)
            
        elif(self.protocol == XYmodem.PROTOCOL_YMODEM):                    
            # ymodem senders expect NAK on EOT 1
            logging.debug("[YMODEM] sending NAK for EOT 1")
            self.transport_stream.write(XYmodem.NAK)
            
            # ymodem senders will send EOT 2
            logging.debug("[YMODEM] waiting for EOT 2")
            sender_byte = self.transport_stream.read(1)
            
            # if EOT, respond with ACK, if not EOT, ignore and exit
            if(ord(sender_byte) == ord(XYmodem.EOT)):                        
                logging.debug("[YMODEM] sending ACK for EOT 2")                        
                self.transport_stream.write(XYmodem.ACK)
                
            else:
                pass
        
        return XYmodem.RECV_EOT
            
            
    
    def rx__read_verify_block(self, receive_buffer, sequence_index, packet_size):
        # read in the block :
        # [SEQ_NUM 1B][SEQ_NUM_C 1B][BLOCK_DATA packet_size B][CRC 2B or CHECKSUM 1B] 
        sequence_num = ord(self.transport_stream.read(1))
        sequence_num_complement = ord(self.transport_stream.read(1))
        receive_buffer.extend(self.transport_stream.read(packet_size))
        if(self.crc_enabled):
            crc_msb = ord(self.transport_stream.read(1))
            crc_lsb = ord(self.transport_stream.read(1))
            crc = (((crc_msb << 8) | crc_lsb) & 0xffff)
        else:
            checksum = ord(self.transport_stream.read(1))
            
        
        # expected values:
        expected_sequence_num = sequence_index % 256
        expected_sequence_num_complement = ((255 - expected_sequence_num) % 256)
        if(self.crc_enabled):
            expected_crc = self.calc_crc16(receive_buffer)
        else:
            expected_checksum = (sum(receive_buffer) + checksum) % 256
        
        
        # integrity checks:
        error_detected = False
        if(len(receive_buffer) != packet_size):
            error_detected = True
            logging.debug("packet contents lost!")
        
        if(sequence_num != expected_sequence_num):
            error_detected = True
            logging.debug("sequence num mismatch, exp: %d got: %d", expected_sequence_num, sequence_num)
            
        if(sequence_num_complement != expected_sequence_num_complement):
            error_detected = True
            logging.debug("sequence num complement mismatch, exp: %d got: %d", expected_sequence_num_complement, sequence_num_complement)
        
        if(self.crc_enabled):
            if(crc != expected_crc):
                error_detected = True
                logging.debug("crc mismatch")
                logging.debug("expected: %x actual: %x",expected_crc, crc)
                
        else:
            if(checksum != expected_checksum):
                error_detected = True
                logging.debug("checksum mismatch")
                logging.debug("expected: %x actual: %x",expected_checksum, checksum)
        
        # return status based on verification
        if(error_detected == True):
            logging.debug("seq: %d ERROR", sequence_index)
            return XYmodem.RECV_FAIL
            
        else:
            logging.debug("block %d verify OK [%d bytes]", sequence_index, len(receive_buffer))
            return XYmodem.RECV_OK
        
    
    def rx__receive_file(self, dir_path, file_name = None, file_size = None):
        if(file_name == None):
            # if file name is not specified, generate one
            time_now = time.strftime("%Y_%m_%d_%H_%M_%S", time.localtime(time.time()))
            logging.debug("%s: %s", "time_now", time_now)
            file_base_name = "uart_rx" + "_" + time_now
            file_extension = ".bin"
            file_name = file_base_name + file_extension
            if(os.path.exists(os.path.join(dir_path,file_name))):
                append_num = 1
                
                while(True):
                    file_name_temp = file_base_name + str(append_num) + file_extension
                    if(not os.path.exists(os.path.join(dir_path,file_name_temp))):
                        # take this as the final file_name
                        file_name = file_name_temp
                        break
                    
                    else:
                        append_num += 1
            
        file_path = os.path.join(dir_path,file_name)
        
        logging.debug("file_path: %s", file_path)
        file_stream = open(file_path, "wb")
        
        sequence_index = 1
        total_bytes = 0
        
        # execute the NCG handshake with the sender
        ncg_response = self.rx__do_ncg_handshake()
        
        if(ncg_response == XYmodem.DATA_SOH):
            packet_size = XYmodem.PACKET_SIZE_128
            
        elif(ncg_response == XYmodem.DATA_STX):
            packet_size = XYmodem.PACKET_SIZE_1024
            
        else:
            # we do not expect any other response here (CANCEL, EOT, TIMEDOUT, UNKNOWN)
            # for all fo these responses, we abort.
            return XYmodem.RX_RECV_FAIL
        
        # if we got data block ncg response (SOH/STX) go to the receive block loop
        retry_count = 0
        while(True):
            # ready an empty receive buffer for the next block
            receive_buffer = bytearray([])
            
            # read the available block and verify:
            receive_status = self.rx__read_verify_block(receive_buffer = receive_buffer, 
                                                            sequence_index = sequence_index,
                                                            packet_size = packet_size)
            
            if(receive_status == XYmodem.RECV_OK):
                # we have got the packet data into the receive buffer we passed in
                total_bytes += packet_size
                sequence_index += 1
                
                if(total_bytes > file_size):
                    valid_packet_size = packet_size - (total_bytes - file_size)
                    receive_buffer = receive_buffer[:valid_packet_size]
                    
                file_stream.write(receive_buffer)
                logging.debug("total bytes received [%d]", total_bytes)
                
                # reset the retry count for successful transfer
                retry_count = 0 
                
                # send ACK for this block:
                logging.debug("send ACK")
                self.transport_stream.write(XYmodem.ACK)
                
            else:
                retry_count += 1
                
                # block read/verify failed!
                logging.debug("SEQ [%d] receive failed, retry_count [%d]", sequence_index, retry_count)
                
                if(retry_count >= self.retry_count_max):
                    # exceeded the threshold for errors, abort now
                    logging.debug("exceeded the retry threshold, abort")
                    return XYmodem.RX_RECV_FAIL
                
                else:
                    # retry getting block from the sender
                    # send NAK for this block:
                    logging.debug("send NAK to request block again")
                    self.transport_stream.write(XYmodem.NAK)
            
            
            # wait for sender for next step 
            # SOH/STX -> block available
            # EOT -> end of transfer, do the EOT sequence
            # CANCEL -> sender cancelled, abort
            # TIMEDOUT/UNKNOWN -> error, abort
            hdr_byte = self.rx__wait_for_hdr()
            
            if(hdr_byte == XYmodem.DATA_SOH):
                packet_size = XYmodem.PACKET_SIZE_128
                continue # read block
            
            elif(hdr_byte == XYmodem.DATA_STX):
                packet_size = XYmodem.PACKET_SIZE_1024
                continue # read block
                
            elif(hdr_byte == XYmodem.DATA_EOT):
                logging.debug("recv EOT")
                file_stream.close()
                self.rx__eot_sequence()                
                return XYmodem.RX_RECV_OK
            
            else:
                # CANCEL, UNKNOWN, abort.
                return XYmodem.RX_RECV_FAIL
                
                
    def rx__process_block_zero(self, receive_buffer):
        
        file_name = None
        file_size = None
        file_modified = None
        file_mode = None    
        
        if(receive_buffer[0] == ord(XYmodem.NUL)):
            logging.debug("all files done, end of batch transfer")                        
        
        else:
            logging.debug("got next incoming file details")
            
            # Valid Next File ->     NAME\x00SIZE(DEC)\x20MODTIME(OCT)\x20MODE(OCT)\x20\x00...\x00
            # End Of Batch    ->     \x00...\x00    
            
            file_details_bytearray_list = receive_buffer.split(XYmodem.NUL)
            
            if(len(file_details_bytearray_list) >= 1):
                file_name_bytes = file_details_bytearray_list[0]
                if(len(file_name_bytes) > 0):
                    file_name = file_name_bytes.decode("ascii")
                
            if(len(file_details_bytearray_list) >= 2): # if we have other details as well:
                file_info_bytes = file_details_bytearray_list[1]                
                file_info_bytearray_list = file_info_bytes.split(XYmodem.SPACE)
                
                # if file_size is present in the details
                if(len(file_info_bytearray_list) >= 1):
                    file_size_bytes = file_info_bytearray_list[0]
                    if(len(file_size_bytes) > 0):
                        file_size = int(file_size_bytes.decode("ascii")) # decimal
                
                # if file_modified is present in the details
                if(len(file_info_bytearray_list) >= 2):    
                    file_modified_bytes = file_info_bytearray_list[1]
                    if(len(file_modified_bytes) > 0):
                        file_modified = int(file_modified_bytes.decode("ascii"), base=8) # octal
                    
                # if file_mode is present in the details                
                if(len(file_info_bytearray_list) >= 3):    
                    file_mode_bytes = file_info_bytearray_list[2]
                    if(len(file_mode_bytes) > 0):
                        file_mode = int(file_mode_bytes.decode("ascii"), base=8) # octal
        
            if(not file_name == None):
                logging.debug("%-16s: %s", "[file_name]", file_name)
                
            if(not file_size == None):
                logging.debug("%-16s: %d B", "[file_size]", file_size)
            
            if(not file_modified == None):
                logging.debug("%-16s: %d", "[file_modified]", file_modified)
            
            if(not file_mode == None):
                logging.debug("%-16s: 0x%x", "[file_mode]", file_mode)
        
        file_details_list = [file_name, file_size, file_modified, file_mode]
        
        return file_details_list
    
    
    def recv(self, dir_path):
        # START of files-loop
        while(True):
            
            file_details_list = [None, None, None, None]
            
            # if we have been given a file_list to receive, then we operate according to the file_list
            # otherwise
            # for ymodem, we wait until we are indicated by the sender that batch send is done using block 0
            # for xmodem, we exit after receiving one file
            
            if(self.protocol == XYmodem.PROTOCOL_YMODEM):
                
                # retry loop block for NCG
                retry_count = 0
                # block 0
                sequence_index = 0
                
                # if block 0 == end of batch, exit
                # else goto receive_file()
                receive_buffer = bytearray([])
                
                # execute the NCG handshake with the sender
                ncg_response = self.rx__do_ncg_handshake()
                
                if(ncg_response == XYmodem.DATA_SOH):
                    packet_size = XYmodem.PACKET_SIZE_128
                    
                elif(ncg_response == XYmodem.DATA_STX):
                    packet_size = XYmodem.PACKET_SIZE_1024
                    
                else:
                    # we do not expect any other response here (CANCEL, EOT, TIMEDOUT, UNKNOWN)
                    # for all fo these responses, we abort.
                    return XYmodem.RX_RECV_FAIL
            
                # START of retry-loop-for-block-zero
                while(True):                    
                    receive_status = self.rx__read_verify_block(receive_buffer = receive_buffer, 
                                                                    sequence_index = sequence_index,
                                                                    packet_size = packet_size)
                    
                    
                    
                    if(receive_status == XYmodem.RECV_OK):
                        # send ACK for this block:
                        logging.debug("send ACK")
                        self.transport_stream.write(XYmodem.ACK)
                        
                        # process the block zero, and get the file details
                        file_details_list = self.rx__process_block_zero(receive_buffer = receive_buffer)
                        
                        file_name = file_details_list[0]
                        if(file_name == None):
                            # end of batch from sender, we are done.
                            return XYmodem.RX_RECV_OK
                        
                        else:
                            # we got the file details of the next incoming file from sender, receive file
                            break # from the retry-loop-for-block-zero to the files-loop
                        
                    else:
                        retry_count += 1
                        
                        # block read/verify failed!
                        logging.debug("SEQ [%d] receive failed, retry_count [%d]", sequence_index, retry_count)
                        
                        if(retry_count >= self.retry_count_max):
                            # exceeded the threshold for errors, abort now
                            logging.debug("exceeded the retry threshold, abort")
                            return XYmodem.RX_RECV_FAIL
                        
                        else:
                            # retry getting block from the sender
                            # send NAK for this block:
                            logging.debug("send NAK to request block again")
                            self.transport_stream.write(XYmodem.NAK)
                    
                    
                    # wait for sender for next step 
                    # SOH/STX -> block available
                    # EOT -> end of transfer, do the EOT sequence
                    # CANCEL -> sender cancelled, abort
                    # TIMEDOUT/UNKNOWN -> error, abort
                    hdr_byte = self.rx__wait_for_hdr()
                    
                    if(hdr_byte == XYmodem.DATA_SOH):
                        packet_size = XYmodem.PACKET_SIZE_128
                        continue # read block
                    
                    elif(hdr_byte == XYmodem.DATA_STX):
                        packet_size = XYmodem.PACKET_SIZE_1024
                        continue # read block
                        
                    elif(hdr_byte == XYmodem.DATA_EOT):
                        self.rx__eot_sequence()
                    else:
                        # EOT, CANCEL, UNKNOWN, abort.
                        return XYmodem.RX_RECV_FAIL
                # END of retry-loop-for-block-zero
                                       
            
            elif( (self.protocol == XYmodem.PROTOCOL_XMODEM) or
                  (self.protocol == XYmodem.PROTOCOL_XMODEM1k) ):                
                # use the file name given in as parameter!
                pass
            
            
            file_name = file_details_list[0]
            file_size = file_details_list[1]
            #file_modified = file_details_list[2]
            #file_mode = file_details_list[3]
            
            # receive the file now
            file_receive_status = self.rx__receive_file(dir_path, file_name, file_size)
            
            # after the file is received, for xmodem protocols, exit the loop
            # for ymodem, continue the loop, the exit is signified by the block zero (end of batch)
            if( (self.protocol == XYmodem.PROTOCOL_XMODEM) or
                  (self.protocol == XYmodem.PROTOCOL_XMODEM1k) ):
                return file_receive_status
            
        # END of files-loop
        
        return
    ###################################        RECEIVER FUNCTIONS --        ###################################
    
    
    ###################################        COMMON FUNCTIONS ++        ###################################    
    def calc_crc16(self,data_bytes, crc = 0):
        for each_byte in bytearray(data_bytes):
            crc = ((crc << 8) ^ XYmodem.crc16table[(((crc >> 8) ^ each_byte) & 0xff)]) & 0xffff
            
        return crc & 0xffff
    
    
    # caution: this will sleep for sometime, to let further junk come in, and then purge
    # this *may* also purge the response byte we are actually waiting for in the caller of
    # this function. As long as the other guy is guaranteed to retry sending the response byte
    # on his end, we should be ok calling this function
    # the amount of time to sleep to wait for junk to accumulate should be decided by the caller
    # this will probably be based on the retry interval on the other guy!
    def purge_transport_stream_input(self, sleep_interval_sec):
        # sleep for some time to let all the junk come in
        time.sleep(sleep_interval_sec)
        
        # purge everything
        self.transport_stream.reset_input_buffer()
    ###################################        COMMON FUNCTIONS --        ###################################
          
        

###################################        MAIN WRAPPER ++        ###################################        
XYMODEM_SYS_EXIT_OK                   = 0
XYMODEM_SYS_EXIT_FAIL                 = 1        

def sig_handler(signum, frame):
    print("CTRL+C!")
    sys.exit()

        
def main(argv=None): # IGNORE:C0111
    '''Command line options.'''

    # logging
    stream_handler = logging.StreamHandler()
    stream_handler_formatter = logging.Formatter("[%(filename)-18s:%(lineno)4d  %(funcName)24s()] %(message)s")    
    stream_handler.setFormatter(stream_handler_formatter)
    stream_handler.setLevel(logging.DEBUG)
    
    file_handler = logging.FileHandler("test.log", "w")
    file_handler_formatter = logging.Formatter('[%(asctime)s %(filename)18s:%(lineno)4s %(funcName)24s()] %(levelname)-8s %(message)s')
    file_handler.setFormatter(file_handler_formatter)
    file_handler.setLevel(logging.DEBUG)
    
    logging.basicConfig(level=logging.DEBUG, handlers = [stream_handler, file_handler])
    
    signal.signal(signal.SIGINT, sig_handler)


    # argument parser
    if argv is None:
        argv = sys.argv
        
    else:
        sys.argv.extend(argv)    
 
    try:
        # setup argument parser
        parser = argparse.ArgumentParser(description = "", formatter_class = argparse.RawDescriptionHelpFormatter)
        
        # tx or rx ?
        parser_group_tx_rx = parser.add_mutually_exclusive_group(required=True)
        parser_group_tx_rx.add_argument("--tx", action="store_true", help="")
        parser_group_tx_rx.add_argument("--rx", action="store_true", help="")
        
        # com port and baud ?
        parser.add_argument("--port", dest="port", help="", required=True)
        parser.add_argument("--baud", dest="baud", help="", type=int, required=True)
        
        # protocol and crc-enabled ?
        parser.add_argument("--protocol", dest="protocol", choices=["xmodem","xmodem1k","ymodem"], help="", required=True)
        parser.add_argument("--crc", action="store_true", help="", required=False)
        
        # for sender, file(s) to send as a list
        parser.add_argument("--file", dest="file", default=None, help="", nargs="+", required=False)
        
        # for receiver we need to specify the destination dir to save the incoming files
        parser.add_argument("--dir", dest="dir", default=None, help="", required=False)
    
        # Process arguments
        args = parser.parse_args()
        
        # additional checking for tx and rx modes:
        if(args.tx == True):
            if(args.file  == None):
                logging.error("tx mode requires files to be sent as a --file argument!")
                return XYMODEM_SYS_EXIT_FAIL
            
            if(not args.dir == None):
                logging.error("tx mode does not expect a --dir argument!")
                return XYMODEM_SYS_EXIT_FAIL
            
        if(args.rx == True):
            if(args.dir == None):
                logging.error("rx mode requires files to be sent as a --dir argument!")
                return XYMODEM_SYS_EXIT_FAIL
            
            if(not args.file == None):
                logging.error("rx mode does not expect a --file argument!")
                return XYMODEM_SYS_EXIT_FAIL
    
        logging.debug("%-24s [%s]", "port", args.port)
        logging.debug("%-24s [%d]", "baud", args.baud)
        logging.debug("%-24s [%s]", "protocol", args.protocol)
        logging.debug("%-24s [%s]", "crc", args.crc)
        
        if(args.protocol=="xmodem"):
            protocol = XYmodem.PROTOCOL_XMODEM
                                    
        elif(args.protocol=="xmodem1k"):
            protocol = XYmodem.PROTOCOL_XMODEM1k
            
        elif(args.protocol=="ymodem"):
            protocol = XYmodem.PROTOCOL_YMODEM        
    
    
        if( (args.tx == True) or 
            (args.rx == True) ):            
            try:                
                ser = serial.Serial(port = args.port,           # port identifier name string
                                    baudrate = args.baud,       # baudrate as integer
                                    timeout = 10,               # read timeout in sec as float (None=blocking, float=nonblocking)
                                    write_timeout = 3)          # write timeout in sec as float (None=blocking, float=nonblocking)
                
            except ValueError as vE:
                logging.debug("%-24s %s",   "ValueError", vE)
                logging.error(traceback.format_exc())
                return XYMODEM_SYS_EXIT_FAIL
            
            except SerialException as sE:
                logging.debug("%-24s %s",   "SerialError", sE)
                logging.error(traceback.format_exc())
                return XYMODEM_SYS_EXIT_FAIL
            
            logging.debug("%-24s %s",   "serial init ok", ser)
            
            # this is not required, unless the pySerial object is instantiated
            # without a "port" name specified!   
            #if(not ser.isOpen()):
            #    ser.open()

            xymodemInstance = XYmodem(protocol = protocol,      # which modem protocol
                                      transport_stream = ser,   # transport stream to use for read/write
                                      crc_enabled = args.crc,   # keep crc enabled (enforced on receiver, but not on sender)
                                      retry_count_max = 3,      # how many times to retry to send or receive indicators(byte/data)
                                      cancel_count_max = 2)     # how many cancels to wait for before aborting
 
    
            t_a = datetime.datetime.now()
            
            if (args.tx == True):                
                logging.debug("%-24s [%s]", "mode", "tx")
                logging.debug("%-24s %s",   "file(s)", args.file)
                
                status = xymodemInstance.send(file_list = args.file)
                
                logging.debug("%-24s %d", "transfer file(s) status", status)
            
            elif(args.rx == True):                
                logging.debug("%-24s [%s]", "mode", "rx")
                logging.debug("%-24s [%s]", "dir", args.dir)
                
                status = xymodemInstance.recv(dir_path = args.dir)
                
                logging.debug("%-24s %d ", "receive file(s) status", status)
                
            t_b = datetime.datetime.now()
            t_diff = t_b - t_a
            
            tddays, rem = divmod(t_diff.total_seconds(), (3600*24))
            tdhours, rem = divmod(rem, 3600)
            tdmin, rem = divmod(rem, 60)
            tdsec = rem
            
            runtime_str = ""
            if(tddays > 0):
                runtime_str += str(int(tddays)) + "d, "
            if(tdhours > 0):
                runtime_str += str(int(tdhours)) + "h, "
            if(tdmin > 0):
                runtime_str += str(int(tdmin)) + "m, "
            runtime_str += str(tdsec) + "s"

            logging.debug("runtime = %s", runtime_str)

            
            ser.close()
            logging.debug("%-24s %s",   "serial close ok", ser)
                        
            return XYMODEM_SYS_EXIT_OK
            
    except KeyboardInterrupt:
        print("ctrl-c, abort.")
        return XYMODEM_SYS_EXIT_FAIL       


    # default exit ok
    return XYMODEM_SYS_EXIT_OK
    

if __name__ == "__main__":
    sys.exit(main())
    
###################################        MAIN WRAPPER --        ###################################
